/* $Id: e2_list.h 874 2008-05-07 14:47:27Z tpgww $

Copyright (C) 2004-2008 tooar <tooar@gmx.net>

This file is part of emelfm2.
emelfm2 is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

emelfm2 is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with emelFM2; see the file GPL. If not, contact the Free Software
Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef __E2_LIST_H__
#define __E2_LIST_H__

#include "emelfm2.h"

gint e2_list_strcmp (const gchar *a, const gchar *b);
void e2_list_update_history (const gchar *command, GList **history,
	guint *cur, guint max, gboolean doubl);
gchar **e2_list_to_strv (GList *list);
GList *e2_list_copy_with_data (GList *list);
void e2_list_free_with_data (GList **list);
//void e2_list_free_data_only (GList **list);
//GList *e2_list_find_data_string (GList *list, gchar *search_text);
void e2_list_nth_break (GList **list, guint breaker, GList **part1, GList **part2);
void e2_list_break (GList **list, GList *breaker, gboolean nullatend, GList **part1,
	GList **part2);

#endif //ndef __E2_LIST_H__
