/* $Id: e2_bookmark.h 874 2008-05-07 14:47:27Z tpgww $

Copyright (C) 2004-2008 tooar <tooar@gmx.net>

This file is part of emelfm2.
emelfm2 is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

emelfm2 is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with emelFM2; see the file GPL. If not, contact the Free Software
Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef __E2_BOOKMARK_H__
#define __E2_BOOKMARK_H__

#include "emelfm2.h"

gboolean e2_bookmark_open (gpointer from, E2_ActionRuntime *art);
void e2_bookmark_add_cb (GtkWidget *widget, gchar *arg);
gboolean e2_bookmark_click_cb (GtkWidget *widget,
	GdkEventButton *event, GtkTreePath* path);

//void e2_bookmark_clean ();
void e2_bookmark_actions_register ();
void e2_bookmark_options_register ();

#endif //ndef __E2_BOOKMARK_H__
